
#include  "EdWws2812.h"
#include <SPI.h>
#include <Wire.h>


EdWws2812::EdWws2812( int liczbaLed, byte cfg, byte csPin, uint32_t spd )
  :    liczbaLed_(liczbaLed)
  ,    cfg_(cfg)
  ,    speed_(spd)
  ,    csPin_(csPin)
{
  //----- Inicjalizacja interfejsu SPI lub I2C -----//
  if ( csPin_ ) { // Jesli CS <> 0 to interfejs I2C
    pinMode(csPin_, OUTPUT);
    SPI.begin();
  }
  else {  // CS = 0 - interfejs I2C
    adrI2c_ = WS2812_BASE_I2C >> 1 ;
    adrI2c_ += cfg_ & 7;  // Adres I2C
    Wire.setClock(spd);
    Wire.begin();
  }

  //----- Rezerwacja wymaganego obszaru RAM -----//
  paleta_ = cfg_ & WS2812_CFG_COLOR_MSK;

  if ( paleta_ == WS2812_CFG_COLOR_218 || paleta_ == WS2812_CFG_COLOR_8 ) {
    pixels_ = malloc( liczbaLed_ );    // Kolor 8-bit
    memset( pixels_, 0, liczbaLed_ );
    EdWws2812_paleta = 8;
    EdWws2812_Ram = liczbaLed_;
  }
  else if ( paleta_ == WS2812_CFG_COLOR_16 ) {
    pixels_ = malloc( liczbaLed_ * 2 );    // Kolor 16-bit
    memset( pixels_, 0, liczbaLed_ * 2 );
    EdWws2812_paleta = 16;
    EdWws2812_Ram = liczbaLed_ * 2;
  }
  else {
    pixels_ = malloc( liczbaLed_ * 3);
    memset( pixels_, 0, liczbaLed_ * 3);
    EdWws2812_paleta = 24;
    EdWws2812_Ram = liczbaLed_ * 3;
  }
  if ( ! pixels_ ) EdWws2812_Err = 1;
}


byte* EdWws2812::begin() {
  return ( EdWws2812_Err );
}


void EdWws2812::setLed( int nrDiody, byte r, byte g, byte b ) {
  byte *p = pixels_;

  if ( pixels_  ) {
    if ( nrDiody < liczbaLed_ ) {
      if ( paleta_ == WS2812_CFG_COLOR_218 ) {  // 8-bit 218 bezpiecznych barw
        p = p + nrDiody;
        *p = r;
      }
      else if ( paleta_ == WS2812_CFG_COLOR_8 ) {  // 8-bit R3G3B2
        // rrrGGGbb
        uint8_t k = ((r & 0b11100000) | ((g >> 3) & 0b00011100) | (b >> 6));
        p = p + nrDiody;
        *p = k;
      }
      else if ( paleta_ == WS2812_CFG_COLOR_16 ) {  // 16-bit
        p = p + nrDiody * 2;
        uint16_t k = (((r >> 3) << 11) | ((g >> 2) << 5) | (b >> 3));
        *p++ = k;
        *p++ = k >> 8;
      }
      else {  // 24-bit
        p = p + nrDiody * 3;
        *p++ = r;
        *p++ = g;
        *p++ = b;
      }
    }
  }
}


void EdWws2812::clr() {
  if ( pixels_ ) {
    if ( paleta_ == WS2812_CFG_COLOR_218 || paleta_ == WS2812_CFG_COLOR_8 ) memset( pixels_, 0 , liczbaLed_ ); // 8-bit
    else if ( paleta_ == WS2812_CFG_COLOR_16 ) memset( pixels_, 0 , liczbaLed_ * 2 ); // 16-bit
    else memset( pixels_, 0 , liczbaLed_ * 3 ); // 24-bit
  }
}


void EdWws2812::show( byte bright ) {
  byte cfg = cfg_ & ~WS2812_CFG_NODISPLAY; // Wycinamy display
  cfg |= paleta_;

  bright &= 0xFC;
  

  if ( csPin_ ) { // Jesli CS <> 0 to interfejs I2C
    if ( speed_ == 8e6 ) SPI.setClockDivider(SPI_CLOCK_DIV2);   // 8MHz za duzo, gubi bajty
    else if ( speed_ == 4e6 ) SPI.setClockDivider(SPI_CLOCK_DIV4);   // 4MHz
    else if ( speed_ == 2e6 ) SPI.setClockDivider(SPI_CLOCK_DIV8);   // 2MHz
    else if ( speed_ == 1e6 ) SPI.setClockDivider(SPI_CLOCK_DIV16);   // 1MHz
    else if ( speed_ == 500000 ) SPI.setClockDivider(SPI_CLOCK_DIV32);   // 500kHz
    else if ( speed_ == 250000 ) SPI.setClockDivider(SPI_CLOCK_DIV64);   // 250kHz
    else if ( speed_ == 125000 ) SPI.setClockDivider(SPI_CLOCK_DIV128);   // 125kHz
    else SPI.setClockDivider(SPI_CLOCK_DIV8);    // 2MHz

    digitalWrite( csPin_, LOW );  // CS = L

    SPI.transfer(cfg);  // Konfiguracja + adres płytki
    SPI.transfer( 0);  // CMD
    SPI.transfer( 0 | bright ); // Adres H od ktorego zapisujemy dane, jasność LED (6 najstarszych bitów)
    SPI.transfer( 0 );  // Adres L od ktorego zapisujemy dane

    if ( pixels_ ) {
      byte *p = pixels_;
      if ( paleta_ == WS2812_CFG_COLOR_218 || paleta_ == WS2812_CFG_COLOR_8 ) {
        for (uint16_t x = 0; x < liczbaLed_; x++) SPI.transfer(*p++);
      }
      else if ( paleta_ == WS2812_CFG_COLOR_16 ) {
        for (uint16_t x = 0; x < liczbaLed_; x++) {
          SPI.transfer(*p++);
          SPI.transfer(*p++);
        }
      }
      else  {
        for (uint16_t x = 0; x < liczbaLed_; x++) {
          SPI.transfer(*p++);
          SPI.transfer(*p++);
          SPI.transfer(*p++);
        }
      }
    }
    digitalWrite( csPin_, HIGH ); // CS = H
  }
  else { // CS = 0 - interfejs I2C
    uint16_t l = 0;
    uint16_t adrLed;
    byte *p = pixels_;
    //while ( l < liczbaLed_  )
    {
      //cfg &= WS2812_CFG_NODISPLAY; // Wycinamy display
      //if ( l < (liczbaLed_ - 8 ) ) cfg &= ~WS2812_CFG_NODISPLAY; // gdy ostatnia ramka danych

      byte adrH = adrLed >> 8;
      byte adrL = adrLed & 0xFF;

      TWI_Master_Initialise( speed_ );

      TWI_Start(); TWI_Write( adrI2c_ << 1 ); //Wire.beginTransmission(adrI2c_); // transmit to device
      TWI_Write( cfg ); //Wire.write( cfg );  // Konfiguracja + adres płytki
      TWI_Write(0);//Wire.write(0);  // CMD
      TWI_Write( adrH | bright ); //Wire.write( adrH | bright );  // Adres H od ktorego zapisujemy dane, jasność LED (6 najstarszych bitów)
      TWI_Write( adrL );//Wire.write( adrL );  // Adres L od ktorego zapisujemy dane
      if ( pixels_ ) {
        //for (byte x = 0; x < 8; x++) {  // 8 * 3 + 4 = 28 (FIFO I2C ma 32 bajty)
        if ( paleta_ == WS2812_CFG_COLOR_218 || paleta_ == WS2812_CFG_COLOR_8 ) {
          for (uint16_t x = 0; x < liczbaLed_; x++) {
            TWI_Write( *p++ );//Wire.write(*p++);
            adrLed ++;
          }
          l++;
        }
        else if ( paleta_ == WS2812_CFG_COLOR_16 ) {
          for (uint16_t x = 0; x < liczbaLed_; x++) {
            TWI_Write( *p++ );//Wire.write(*p++);
            TWI_Write( *p++ );//Wire.write(*p++);
            adrLed += 2;
          }
          l++;
        }
        else  {
          for (uint16_t x = 0; x < liczbaLed_; x++) {
            TWI_Write( *p++ );//Wire.write(*p++);
            TWI_Write( *p++ );//Wire.write(*p++);
            TWI_Write( *p++ );//Wire.write(*p++);
            adrLed += 3;
          }
          l++;
        }
      }
      TWI_Stop(); EdWws2812_i2c = TWI_Status; //EdWws2812_i2c = Wire.endTransmission();    // stop transmitting
      delay(3);
      /*
        char txt[50];
        sprintf( txt, " p=%d %x %02x", adrLed, adrH, adrL ); Serial.print(txt);
      */
    }
  }
}


#define nop()  __asm__ __volatile__( "nop" )   // Wstawia rozkaz NOP do kodu

void EdWws2812::SPI_Send8MHz( byte dane ) {
  SPDR = dane;
  nop(); nop(); nop(); nop();
  nop(); nop(); nop(); nop();
  nop(); nop(); nop(); nop();
  nop(); nop(); nop(); nop();
}


//----------------------------------------------------------------------//
void EdWws2812::TWI_Master_Initialise( uint32_t speed )
{
  //TWI_Master_Off();

  //SCL frequency CPU Clock frequency / ( 16 + 2(TWBR) + 4 · TWPS ) TWPS = PRESKALER
  //F_CPU/(16+2*TWBR+4*(TWSR&3))
  // TWBR min 10
  speed /= 1000;
  TWBR = ((F_CPU / 1000 / speed) - 16) / 2; // TWISCK - CLK w kHz
  /*
    #define TWBR_100KHZ   ((F_CPU/1000/100)-16)/2;
    #define TWBR_400KHZ   ((F_CPU/1000/400)-16)/2;
     #if (((F_CPU/1000/TWISCK)-16)/2) < 10
    //    #error "Podobno sieje bledy"
      #warning "Podobno sieje bledy"
    //  if ( TWBR < 10 ) TWBR=10;
    #endif
  */
  // TWSR = TWI_TWPS;   // Not used. Driver presumes prescaler to be 00.
  //  TWDR = 0xFF;      // Default content = SDA released.
  TWCR = (1 << TWEN); // Enable TWI-interface and release TWI pins.
}


byte EdWws2812::TWI_Action( byte command ) {
  byte overtime = 100;


  //write command to TWCR and make sure TWINT is set
  TWCR = (command | (1 << TWINT) | (1 << TWEN) );
  //now wait for TWINT to be set again (when the operation is completed)
  while ( !(TWCR & (1 << TWINT)) )
  {
    _delay_us(1);
    overtime--;
    if ( !overtime ) break;
  }
  /*
    There are two status codes that do not correspond to a defined TWI state, see Table
    120.
    Status 0xF8 indicates that no relevant information is available because the TWINT Flag
    is not set. This occurs between other states, and when the TWI is not involved in a serial
    transfer.
    Status 0x00 indicates that a bus error has occurred during a 2-wire Serial Bus transfer.
    A bus error occurs when a START or STOP condition occurs at an illegal position in the
    format frame. Examples of such illegal positions are during the serial transfer of an
    address byte, a data byte, or an acknowledge bit. When a bus error occurs, TWINT is
    set. To recover from a bus error, the TWSTO Flag must set and TWINT must be cleared
    by writing a logic one to it. This causes the TWI to enter the not addressed Slave mode
    and to clear the TWSTO Flag (no other bits in TWCR are affected). The SDA and SCL
    lines are released, and no STOP condition is transmitted.
  */
  /*
    Istnieją dwa kody stanu, które nie odpowiadają określonym stanie TWI, patrz tabela 120.

    Stan 0xF8 wskazuje, że żadne istotne informacje są dostępne ponieważ TWINT Flag
    nie jest ustawiona. Dzieje się tak gdy TWI nie jest zaangarzowany w transfer.
    Jeśli status F8 pojawia się po wysłaniu warunku statt, może to oznaczać uszkodzenie magistrali IIC (zwarcie lub brak rezystorów podciągających).

    Stan 0x00 oznacza, że wystąpił błąd magistrali IIC.
    Błąd występuje, gdy START lub STOP pojawił się w nieodpowiednim miejscu najczęściej podczas nielegalnego przejęcia magistrali.
    Gdy wystąpi błąd magistrali (w trybie Slave), należy wyserować TWSTO i ustawiC TWINT.
    TWI nie wygeneruje nic na magistrali ale zwolni linie.
  */
  TWI_Status = TWSR & 0b11111100;
  return ( TWI_Status );
}


byte EdWws2812::TWI_Stop() {
  byte overtime = 100;


  TWCR = (1 << TWINT) | (1 << TWEN) | (1 << TWSTO);
  // Generując stop czekamy na bit TWSTO a nie WTINT
  while ( (TWCR & (1 << TWSTO)) )
  {
    _delay_us(1);
    overtime--;
    if ( !overtime ) break;
  }


  return ( TWSR );
}


byte EdWws2812::TWI_StatusTransceiver_Busy() {
  return ( TWCR & (1 << TWIE) );  // IF TWI Interrupt is enabled then the Transceiver is busy
}


byte EdWws2812::TWI_Start() {
  return ( TWI_Action( 1 << TWSTA ) );
}


byte EdWws2812::TWI_Read( byte ack ) {
  if ( ack ) TWI_Action( 1 << TWEA );
  else TWI_Action( 0 );

  TWI_Status = TWSR;
  return ( TWDR );
}


byte EdWws2812::TWI_Write( byte d ) {
  TWDR = d;
  return ( TWI_Action( 0 ) );
}


